# pylint: disable=unused-argument
import json
import os
import tempfile
from pathlib import Path

import mock
from dagster import Array, Field
from dagster.core.test_utils import environ
from dagster_cloud.agent.cli import app
from dagster_cloud.auth.constants import DEPLOYMENT_NAME_HEADER
from dagster_cloud.workspace.docker import DockerUserCodeLauncher
from typer.testing import CliRunner


def _check_dagster_home(expected: str):
    def _check_inner():
        assert os.getenv("DAGSTER_HOME") == expected

    return _check_inner


def test_run_command_with_env():
    runner = CliRunner()

    with tempfile.TemporaryDirectory() as temp_dir:
        with environ({"DAGSTER_HOME": temp_dir}):
            with mock.patch(
                "dagster_cloud.agent.cli.run_local_agent", _check_dagster_home(temp_dir)
            ):
                result = runner.invoke(app, [])

                assert result.exit_code == 0, result.output + " : " + str(result.exception)


def test_run_command_with_no_home():
    runner = CliRunner()

    with environ({"DAGSTER_HOME": None}):
        result = runner.invoke(app, [])

        assert result.exit_code == 1, result.output + " : " + str(result.exception)
        assert "No directory provided" in result.output


def test_run_command_with_argument():
    runner = CliRunner()

    # Test param works
    with tempfile.TemporaryDirectory() as temp_dir:
        with mock.patch(
            "dagster_cloud.agent.cli.run_local_agent",
            _check_dagster_home(str(Path(temp_dir).resolve())),
        ):
            result = runner.invoke(app, [temp_dir])

            assert result.exit_code == 0, result.output + " : " + str(result.exception)

    # Test param overrides DAGSTER_HOME env var
    with tempfile.TemporaryDirectory() as temp_dir, tempfile.TemporaryDirectory() as temp_dir_env:
        with environ({"DAGSTER_HOME": temp_dir_env}):
            with mock.patch(
                "dagster_cloud.agent.cli.run_local_agent",
                _check_dagster_home(str(Path(temp_dir).resolve())),
            ):
                result = runner.invoke(app, [temp_dir])

                assert result.exit_code == 0, result.output + " : " + str(result.exception)


TEST_TOKEN = "agent:hooli:1b2c814969d746b09d9k4d7e6g065cn4"


def test_run_ephemeral():
    runner = CliRunner()

    # Both directory and ephemeral config provided (error)
    with tempfile.TemporaryDirectory() as temp_dir:
        result = runner.invoke(app, [temp_dir, "--agent-token", TEST_TOKEN, "--deployment", "prod"])

        assert result.exit_code == 1, result.output + " : " + str(result.exception)
        assert "Cannot supply both" in result.output

    # Missing part of ephemeral config (error)
    with tempfile.TemporaryDirectory() as temp_dir:
        result = runner.invoke(app, ["--agent-token", TEST_TOKEN])

        assert result.exit_code == 1, result.output + " : " + str(result.exception)
        assert "both an agent token and a deployment" in result.output

    # Grab the DagsterCloudAgentInstance
    call_instance = [None]

    def mock_run_loop(_agent, instance, *args, **kwargs):
        call_instance[0] = instance

    # Minimal ephemeral config
    with mock.patch("dagster_cloud.agent.cli.DagsterCloudAgent.run_loop", mock_run_loop):
        result = runner.invoke(app, ["--agent-token", TEST_TOKEN, "--deployment", "staging"])

        assert result.exit_code == 0, result.output + " : " + str(result.exception)

        # Ensure DagsterCloudAgentInstance configured correctly
        assert call_instance[0].dagster_cloud_agent_token == TEST_TOKEN
        assert call_instance[0].dagster_cloud_api_headers.get(DEPLOYMENT_NAME_HEADER) == "staging"
        assert call_instance[0].dagster_cloud_url == "https://hooli.agent.dagster.cloud"
        assert call_instance[0].dagster_cloud_api_agent_label == None

    # With agent label
    with mock.patch("dagster_cloud.agent.cli.DagsterCloudAgent.run_loop", mock_run_loop):
        result = runner.invoke(
            app,
            ["--agent-token", TEST_TOKEN, "--deployment", "prod", "--agent-label", "My Test Agent"],
        )

        assert result.exit_code == 0, result.output + " : " + str(result.exception)

        assert call_instance[0].dagster_cloud_agent_token == TEST_TOKEN
        assert call_instance[0].dagster_cloud_api_headers.get(DEPLOYMENT_NAME_HEADER) == "prod"
        assert call_instance[0].dagster_cloud_url == "https://hooli.agent.dagster.cloud"
        assert call_instance[0].dagster_cloud_api_agent_label == "My Test Agent"

    # Ensure they're required, to test supplying config via CLI arg
    def mock_config_type():
        return {
            "networks": Field(Array(str), is_required=True),
            "env_vars": Field(
                [str],
                is_required=True,
                description="The list of environment variables names to forward to the docker container",
            ),
        }

    # With custom user code launcher
    with mock.patch(
        "dagster_cloud.agent.cli.DagsterCloudAgent.run_loop", mock_run_loop
    ), mock.patch(
        "dagster_cloud.workspace.docker.DockerUserCodeLauncher.config_type", mock_config_type
    ):
        result = runner.invoke(
            app,
            [
                "--agent-token",
                TEST_TOKEN,
                "--deployment",
                "prod",
                "--agent-label",
                "My Test Agent",
                "--user-code-launcher",
                "dagster_cloud.workspace.docker.DockerUserCodeLauncher",
                "--user-code-launcher-config",
                json.dumps(
                    {
                        "networks": ["abc", "def"],
                        "env_vars": ["PATH", "DAGSTER_HOME"],
                    }
                ),
            ],
        )

        assert result.exit_code == 0, result.output + " : " + str(result.exception)

        assert call_instance[0].dagster_cloud_agent_token == TEST_TOKEN
        assert call_instance[0].dagster_cloud_api_headers.get(DEPLOYMENT_NAME_HEADER) == "prod"
        assert call_instance[0].dagster_cloud_url == "https://hooli.agent.dagster.cloud"
        assert isinstance(call_instance[0].user_code_launcher, DockerUserCodeLauncher)
        assert call_instance[0].user_code_launcher.env_vars == ["PATH", "DAGSTER_HOME"]



import json
import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

import yaml
from dagster.core.errors import DagsterHomeNotSetError
from dagster.utils.interrupts import capture_interrupts
from dagster.utils.log import default_date_format_string, default_format_string
from dagster_cloud.agent.dagster_cloud_agent import DagsterCloudAgent
from dagster_cloud.instance import DagsterCloudAgentInstance
from typer import Argument, Option, Typer

from ...cli import ui

app = Typer(help="Interact with the Dagster Cloud agent.")


def agent_home_exception():
    dagster_home_loc = (
        (
            f"No Dagster config provided in specified directory {os.getenv('DAGSTER_HOME')}. "
            "You must specify the location of a directory containing a dagster.yaml "
            "file as a parameter or by setting the DAGSTER_HOME environment variable."
        )
        if os.getenv("DAGSTER_HOME")
        else (
            "No directory provided or DAGSTER_CLOUD environment variable set. "
            "You must supply the location of a directory containing a dagster.yaml "
            "file as a parameter or by setting the DAGSTER_HOME environment variable."
        )
    )
    return ui.error(f"No Dagster config found.\n\n{dagster_home_loc}")


def run_local_agent():
    try:
        with DagsterCloudAgentInstance.get() as instance:
            user_code_launcher = instance.user_code_launcher
            user_code_launcher.start()

            with DagsterCloudAgent() as agent:
                agent.run_loop(instance, user_code_launcher, agent_uuid=instance.instance_uuid)
    except DagsterHomeNotSetError:
        raise agent_home_exception()


def run_local_agent_in_environment(dagster_home: Optional[Path]):
    with capture_interrupts():
        try:
            old_env = dict(os.environ)
            if dagster_home:
                os.environ["DAGSTER_HOME"] = str(dagster_home.resolve())
            run_local_agent()
        finally:
            os.environ.clear()
            os.environ.update(old_env)


DAGSTER_YAML_TEMPLATE = """
instance_class:
  module: dagster_cloud.instance
  class: DagsterCloudAgentInstance

dagster_cloud_api:
  agent_token: {token}
  deployment: {deployment}
  {agent_label_entry}

user_code_launcher:
  module: {user_code_launcher_module}
  class: {user_code_launcher_class}
  {user_code_launcher_config_entry}
"""


def run_local_agent_in_temp_environment(
    agent_token: str,
    deployment: str,
    agent_label: Optional[str],
    user_code_launcher_module: str,
    user_code_launcher_class: str,
    user_code_launcher_config: Optional[str],
):
    with TemporaryDirectory() as d:
        config_entry = ""
        if user_code_launcher_config:
            try:
                unindented_config_entry = yaml.dump(
                    {"config": json.loads(user_code_launcher_config)}
                )
                config_entry = unindented_config_entry.replace("\n", "\n  ")
            except json.JSONDecodeError as e:
                raise ui.error(f"Invalid User Code Launcher config JSON:\n{e}")

        with open(os.path.join(d, "dagster.yaml"), "w") as f:
            f.write(
                DAGSTER_YAML_TEMPLATE.format(
                    token=agent_token,
                    deployment=deployment,
                    agent_label_entry=f"agent_label: {agent_label}" if agent_label else "",
                    user_code_launcher_module=user_code_launcher_module,
                    user_code_launcher_class=user_code_launcher_class,
                    user_code_launcher_config_entry=config_entry,
                )
            )
        run_local_agent_in_environment(Path(d))


PROCESS_USER_CODE_LAUNCHER = "dagster_cloud.workspace.user_code_launcher.ProcessUserCodeLauncher"


@app.command(
    help=(
        "Runs the Dagster Cloud agent. The agent can either be run ephemerally by specifying an agent token "
        "and deployment name as CLI options, or the agent can pull its config from a dagster.yaml file. "
        "To use a dagster.yaml file, either pass a directory containing the file as a CLI argument or "
        "set the DAGSTER_HOME environment variable."
    ),
    short_help="Run the Dagster Cloud agent.",
)
def run(
    dagster_home: Optional[Path] = Argument(None),
    agent_token: str = Option(
        None, "--agent-token", "-a", help="Agent token, if running ephemerally."
    ),
    deployment: str = Option(
        None, "--deployment", "-d", help="Deployment, if running ephemerally."
    ),
    agent_label: str = Option(
        None, "--agent-label", "-l", help="Optional agent label, if running ephemerally."
    ),
    user_code_launcher: str = Option(
        PROCESS_USER_CODE_LAUNCHER,
        "--user-code-launcher",
        help="User Code Launcher to use. Defaults to the local Process User Code Launcher.",
    ),
    user_code_launcher_config: str = Option(
        None,
        "--user-code-launcher-config",
        help="Config to supply the User Code Launcher, in JSON format.",
    ),
):
    logging.basicConfig(
        level=logging.INFO,
        format=default_format_string(),
        datefmt=default_date_format_string(),
        handlers=[logging.StreamHandler()],
    )
    if (
        agent_token
        or deployment
        or agent_label
        or user_code_launcher_config
        or user_code_launcher != PROCESS_USER_CODE_LAUNCHER
    ):
        if not agent_token or not deployment:
            raise ui.error("To run ephemerally, must supply both an agent token and a deployment.")
        if dagster_home:
            raise ui.error("Cannot supply both a dagster home directory and ephemeral parameters.")

        user_code_launcher_module = ""
        user_code_launcher_class = user_code_launcher
        if "." in user_code_launcher:
            user_code_launcher_module, user_code_launcher_class = user_code_launcher.rsplit(".", 1)
        run_local_agent_in_temp_environment(
            agent_token,
            deployment,
            agent_label,
            user_code_launcher_module,
            user_code_launcher_class,
            user_code_launcher_config,
        )
    else:
        run_local_agent_in_environment(dagster_home)
